'-------------------------------------------------------------------------------------------
' RogueLike using Random Maze Generator Subroutine
' v1.0 by Womble 27 October 2020
' https://www.thebackshed.com/forum/ViewTopic.php?FID=16&TID=12987#157644
'
' Implementation of Eller's Algorithm:
'   Written by Vegipete (vegipete@gmail.com)
'   June 2020
' Subroutined and Modified by Womble 27 October 2020
' https://www.thebackshed.com/forum/ViewTopic.php?FID=16&TID=12209
'
' Eller's algorithm generates pure simply connected mazes
' of arbitrary length.
' Horizontal/Vertical bias: 
' .5 = even, lower = more horizontal,  higher = more vertical
' By introducing a deliberate 'bug', mazes with loops can
' be generated.
'
' RogueLike Elements of this program based on:
' Golfed Rogue 
' https://codegolf.stackexchange.com/questions/7074/build-a-simple-roguelike
' also https://github.com/nrkn/SimpleRLGolf
' from a rec.games.roguelike.development post
' DLosc QBASIC solution posted on StackExchange on 21APL2015
' licenced under https://creativecommons.org/licenses/by-sa/3.0/
' modified and commented for CMM2 by Womble 26OCT2020
'
'-------------------------------------------------------------------------------------------


'-------------------------------------------------------------------------------------------
' OPTIONS AND DECLARATIONS
'-------------------------------------------------------------------------------------------

'Recommended Program Practice
OPTION EXPLICIT          ' Variables must be typed
OPTION DEFAULT NONE      ' Variables must be declared

'Set array option base to index from 0
OPTION BASE 0

'Start with a clear screen
CLS

'Global Variables and Constants
DIM INTEGER WIDTH%         ' Number of Maze Columns (minimum 5) will be doubled+1 size
DIM INTEGER LENTH%         ' Number of Maze Rows (minimum 5) will be doubled+1 size
DIM INTEGER LOOPS%         ' Allow loops. 0 = NO loops, 1 = loops permited
DIM  FLOAT BIAS = 0.5      ' .5 = even, lower = more horizontal,  higher = more vertical
DIM INTEGER FUNKYBIAS%     ' Funky variable bias flag

'Get Maze Values before defining some constants
GetMazeSettigs

CONST mapR = 2*LENTH%+1    ' Binary Map Rows
CONST mapC = 2*WIDTH+1     ' Binary Map Columns
DIM FLOAT wall(mapR,MapC)  ' Binary Wall Map Array, initially everything is 0 a wall
DIM FLOAT SplatLocR        ' StartPoint Row, also current location of @
DIM FLOAT SplatLocC        ' StartPoint Column, also current location of @
DIM FLOAT ExitLocR         ' StartPoint Row, also current location of X (maze exit)
DIM FLOAT ExitLocC         ' StartPoint Column, also current location of X (maze exit)

'Declare Global Arrays
DIM m$(2*LENTH%+1) length 2*WIDTH+1   ' Maze String Array, each string 2*WIDTH+1 chars long
DIM FLOAT w(WIDTH)
DIM FLOAT p(WIDTH)
DIM FLOAT cnt(WIDTH)

'General Purpose Variables using in main program
DIM a$                     ' Keyboard input string
DIM FLOAT r                ' row location
DIM FLOAT c                ' column location
DIM FLOAT k                ' Keyboard ASCII value


'-------------------------------------------------------------------------------------------
' MAIN PROGRAM CODE
'-------------------------------------------------------------------------------------------

'Print Title and Maze Description
CLS
COLOUR RGB(YELLOW),RGB(BLACK)
gLOCATE 18,50
PRINT "RogueLike Ellers Random Maze Generator"
COLOUR RGB(YELLOW),RGB(GRAY)
gLOCATE 19,50
PRINT "program by Womble and Ellers code by VegiPete"
gLOCATE 21,50
COLOUR RGB(WHITE),RGB(BLACK)
PRINT "Width:" WIDTH% " Length:" LENTH% "  Bias:" BIAS! " Loops: ";
IF LOOPS% THEN PRINT "Y" ELSE PRINT "N";
gLOCATE 22,50
PRINT "Funky: "; 
IF FUNKYBIAS% THEN PRINT "Y" ELSE PRINT "N";

'Print the user instructions
gLOCATE 23,50
PRINT "Arrow Keys to Move, TAB or ESC to exit"

GenEllersMaze             ' Generate the random maze
BuildBinaryWallMap        ' Load maze into binary wall map
prnMaze                   ' Draw the Maze onscreen

r = SplatLocR             ' iitialise current location for @ row
c = SplatLocC             ' iitialise current location for @ column
wall(ExitLocR,ExitLocC) = 0 ' initialise the exit location to 0 (lock the door)

MainLoop:                 ' Main Animation Loop Label
DO
   gLOCATE SplatLocR,SplatLocC
   COLOUR RGB(RED) : PRINT "@"     'Draw the @ character
   a$=""                 'Clear keyboard input string
   do while a$ = ""
      a$ = INKEY$        'Wait for keyboard input
   loop

   k=ASC(RIGHT$(a$,1))   'Keypress up 128, down 239, left 130, right 131 or TAB 9 ESC 27

   'Evaluate Keypress to determine if destination is a wall or a space
   c=SplatLocC+(k=131)-(k=130)  'Where are we trying to move to?
   r=SplatLocR+(k=129)-(k=128)  'current location +1 +0  k=up? + k=down?  truth test

   gLOCATE SplatLocR,SplatLocC  'locate cursor to old @ location
   PRINT " "                    'erase the old @ character

   IF wall(r,c) THEN            'if new location is 1 a space ie. TRUE
      SplatLocR=r               'update the location for the @ character 
      SplatLocC=c               'otherwise location remains the same
   END IF

   ' Check if we have reached the Exit Door
   IF r = ExitLocR AND c = ExitLocC THEN 
      COLOUR RGB(GREEN),RGB(GRAY)
      glocate 25,50
      PRINT "Congratulations You Have Exited the Maze";  'No Carriage Return
      gLOCATE 26,50
      COLOUR RGB(GREEN),RGB(BLACK)
      PRINT "press TAB or ESC to exit the program";      'No Carriage Return
      ' We have reached the Exit Door
   ENDIF

LOOP UNTIL k=9 OR k=27       'TAB or ESC key was pressed so exit

' Exit Program Gracefully
COLOUR RGB(WHITE),RGB(BLACK)
CLS
END



'-------------------------------------------------------------------------------------------
' SUBROUTINES
'-------------------------------------------------------------------------------------------


'-------------------------------------------------------------------------------------------
' GenEllersMaze Subroutine
' Implementation of Eller's Algorithm:
'   Written by Vegipete (vegipete@gmail.com) June 2020
' Subroutined and Modified by Womble 27 October 2020
' https://www.thebackshed.com/forum/ViewTopic.php?FID=16&TID=12209
'-------------------------------------------------------------------------------------------

SUB GenEllersMaze
   LOCAL INTEGER i                     ' Generic Counter Variables
   LOCAL INTEGER j
   LOCAL INTEGER rc
   LOCAL INTEGER sn
   LOCAL INTEGER dd
   LOCAL INTEGER row                   ' Maze Array Row
   LOCAL lp$
   LOCAL rp$

   'Generate the Maze Array m$()
   m$(1) = "#"                         ' Initialise the Map String Array
   m$(2) = "#"

   FOR i = 1 TO WIDTH                  ' Build first Maze element
      m$(1) = m$(1) + "##"             ' Top of maze, all wall
      m$(2) = m$(2) + " #"             ' Second row alternating space and wall
   NEXT i

   m$(3) = m$(1)                       ' Fill in the rest of the Maze elements
   FOR i = 4 TO LENTH*2 Step 2
      m$(i) = m$(2)
      m$(i+1) = m$(3)
   NEXT i

   row = 2                             ' Start building from second row

   FOR i = 1 TO WIDTH : w(i) = i : NEXT i

   'Build the Maze
   Do
      ' do horizontal cell joining
      FOR i = 1 TO WIDTH : p(i) = 0 : NEXT i

      ' Check FUNKBIAS% flag for an interesting variable bias effect
      ' BIAS is a global not a constant so we can do this
      IF FUNKYBIAS% THEN BIAS = row / LENTH / 2

   FOR i = 2 TO WIDTH
      IF Rnd(1) > BIAS THEN
         IF w(i) <> w(i-1) THEN
            ' change set
            IF LOOPS THEN
               w(i) = w(i-1)
            ELSE
               rc = w(i)
               FOR j = 1 TO WIDTH
                  IF w(j) = rc THEN w(j) = w(i-1)
               NEXT j
            ENDIF

             ' update maze
             lp$ = Left$(m$(row),2*(i-1))
             rp$ = Right$(m$(row),2*WIDTH-2*(i-1))
             m$(row) = lp$ + " " + rp$
          ENDIF
       ENDIF
    NEXT i

   ' reorder set ids and count
   FOR i = 1 TO WIDTH : cnt(i) = 0 : w(i) = w(i) + WIDTH : NEXT i
   rc = 1
   FOR i = 1 TO WIDTH
      IF w(i) > WIDTH THEN
         sn = w(i)
         w(i) = rc
         FOR j = i TO WIDTH
            IF w(j) = sn THEN w(j) = rc
         NEXT j
         rc = rc + 1
      ENDIF
      cnt(w(i)) = cnt(w(i)) + 1
   NEXT i

   row = row + 1

   ' do vertical cell joining
   FOR i = 1 TO WIDTH : p(i) = 0 : NEXT i

   FOR i = 1 TO WIDTH
      dd = 0
      IF cnt(w(i)) = 1 THEN
         ' set with single element MUST open downward
         p(i) = w(i)
         dd = 1
      ELSEIF cnt(w(i)) > 1 THEN
         ' set with multi elements randomly opens downward
         IF Rnd(1) < BIAS THEN
            p(i) = w(i)
            dd = 1
         ELSE
            ' each set MUST have at least one downward opening
            cnt(w(i)) = cnt(w(i)) - 1
         ENDIF
      ENDIF
      IF dd THEN
         ' update maze - make vertical passages
         lp$ = Left$(m$(row),2*i-1)
         rp$ = Right$(m$(row),2*(WIDTH-i)+1)
         m$(row) = lp$ + " " + rp$
      ENDIF
   NEXT i

   row = row + 1

   ' build new row from downward passages
   FOR i = 1 TO WIDTH : w(i) = p(i) : NEXT i

   ' fill in new sets
   rc = WIDTH
   FOR i = 1 TO WIDTH
      IF w(i) = 0 THEN
         w(i) = rc
         rc = rc - 1
      ENDIF
   NEXT i

   LOOP WHILE row < 2*LENTH

   ' do final horizontal joining
   FOR i = 2 TO WIDTH
      IF w(i) <> 0 THEN
         IF w(i) <> w(i-1) THEN
            ' change set
            FOR j = i + 1 TO WIDTH
               IF w(j) = w(i) THEN w(j) = 0
            NEXT j
            w(i) = w(i-1)
            ' update maze
            lp$ = Left$(m$(row),2*(i-1))
            rp$ = Right$(m$(row),2*WIDTH-2*(i-1))
            m$(row) = lp$ + " " + rp$
         ENDIF
      ENDIF
   NEXT i

   'Define Maze Entrance and Exit
   SetEndPoints

END SUB


'-------------------------------------------------------------------------------------------
' BuildBinaryWallMap  from m$ Map Strig
'-------------------------------------------------------------------------------------------

SUB BuildBinaryWallMap
   LOCAL FLOAT ii
   LOCAL FLOAT jj
   LOCAL FLOAT tv        ' Binary value 1 (space) or 0 (wall)
   LOCAL STRING sc$
   LOCAL STRING srow$
   FOR ii = 1 to mapR%
      srow$ = m$(ii)
      FOR jj = 1 to mapC%
         sc$ = MID$(srow$,jj,1)
         SELECT CASE sc$
            CASE " " : tv = 1     ' Corridor Space
            CASE "#" : tv = 0     ' Wall
            CASE ELSE : tv = 1    ' Anything Else is a Corridor Space
         END SELECT
         wall(ii,jj) = tv         ' Load value into Global array wall(row,col)
      NEXT jj
   NEXT ii

END SUB


'-------------------------------------------------------------------------------------------
' prnMaze   Draw the Maze onscreen using print statements
'-------------------------------------------------------------------------------------------
SUB prnMaze
   LOCAL wt$                   ' SELECT CASE workaround variable
   LOCAL mR%                   ' Counter for printing a row of the maze
   LOCAL mC%                   ' Counter for individual map character from string

   'Display the m$() array contents
   COLOUR RGB(BROWN)
   glocate 1,1
   FOR mR% = 1 to (LENTH% + LENTH% +1)    ' Maze String Row
      FOR mC% = 1 to 2*WIDTH%+1           ' Maze String Column
         wt$ = MID$(m$(mR%), mC%, 1)      ' Exctract a single character
            SELECT CASE wt$                   ' Test character and select colour
               CASE "@"                       ' Start Position
                  COLOUR RGB(RED),RGB(BLACK)
               CASE "X"                       ' Maze Exit
                  COLOUR RGB(GREEN),RGB(BLACK)
               CASE "#"                       ' Wall
                  COLOUR RGB(BROWN),RGB(GRAY)
               CASE ELSE
                  COLOUR RGB(WHITE),RGB(GRAY) ' Default Colour for anything else
            END SELECT
         PRINT wt$;                        ' Print Character (supress carriage return)
      NEXT mC%                            ' Next Maze String Column (character from string)
      PRINT                               ' Carriage Return at end of row
   NEXT mR%                               ' Next Maze String Row (array item)
   PRINT
   COLOUR RGB(WHITE),RGB(BLACK)           ' Set Colour back to default
   PRINT "Rows " (LENTH% + LENTH% + 1)
   PRINT "Columns " (WIDTH% + WIDTH% + 1)

END SUB


'-------------------------------------------------------------------------------------------
' SetEndPoints subroutine to create the start and endpoints
'-------------------------------------------------------------------------------------------

SUB SetEndPoints
   LOCAL INTEGER i=2                  'Start row for @ character
   LOCAL INTEGER j=2                  'Start column for @ character
   LOCAL INTEGER k=2                  'End row for X exit door
   LOCAL INTEGER l=2                  'END column for X exit door
   LOCAL INTEGER py
   
   ' Door on right side
   py = INT(RND() * LENTH%) + 1       'Exit is on a Random Row
   py = py * 2

   MID$(m$(py),LEN(m$(py))) = "X"     ' Maze EndPoint (Exit)
   k = py                             ' Random Row 
   l = 2*WIDTH%+1                     ' Right Most Column
   ExitLocR = k                       ' Gloval Exit Location 
   ExitLocC = l
 
   ' Door on left side
   py = INT(RND() * LENTH%) + 1       ' Entrance is on a Random Row
   py = py * 2

   MID$(m$(py),1) = "@"               ' Maze StartPoint (Entrance)
   i = py                             ' Randowm Row
   j = 1                              ' Left Most Column
   SplatLocR = i                      ' Gloval Start Position
   SplatLocC = j

END SUB


'-------------------------------------------------------------------------------------------
' gLOCATE subroutine to emulate the LOCATE command in GWBasic nad QBasic
'--------------------------------------------------------------------------=========---=----

SUB gLOCATE row%, col%
  LOCAL x% = col% -1                                           'Columns start at 1 
  LOCAL y% = row% -1                                           'Rows start at 1
  IF x% < 0 OR x% > ( MM.HRES \ MM.INFO(FONTWIDTH) ) THEN      'max character in row
    x% = MM.HRES \ MM.INFO(FONTWIDTH)                          'set to max characters
  ENDIF
  IF y% < 0 OR y% > ( MM.VRES \ MM.INFO(FONTHEIGHT) ) THEN     'max rows onscreen
    y% = MM.VRES \ MM.INFO(FONTHEIGHT)                         'set to max rows
  ENDIF
  PRINT @((MM.INFO(FONTWIDTH)*x%),(MM.INFO(FONTHEIGHT)*y%))""; 'position cursor
  'where x is the character column, and y is the character row
  'https://www.thebackshed.com/forum/ViewTopic.phpPRINT TID=12778&PID=155313#155313#155313
END SUB


'-------------------------------------------------------------------------===--------==-----
'
' PrnDecimal2Binary on the CMM2, using 64bit signed integers 
' based on code from https://www.qb64.org/wiki/Binary
'
'-------------------------------------------------------------------------=====-------------

SUB PrnDecimal2Binary number%, bits%
  '
  ' Prints Binary bits of a Decimal Number. 
  ' Input value number% is a signed 64bit integer
  ' Input value bits% is the number of bits (length) to print
  '
  LOCAL Bin$ = ""                  ' clear the output strings
  LOCAL Binary$ = ""
  LOCAL remain%

  IF bits% = 0 THEN EXIT SUB       ' Nothing to print
  IF number% = 0 THEN EXIT SUB

  DO                               ' Remainder can only be 1 or 0 in each loop. 
    remain% = ABS(number% MOD 2)   ' Make unsigned, remainder is used to create binary number
    number% = number% \ 2          ' Move up one exponent of 2 with integer division
    Bin$ = STR$(remain%)           ' Make remainder a string number
    Binary$ = Bin$ + Binary$       ' Concatenate remainder to binary number string
  LOOP UNTIL number% = 0

  PRINT "  =  ";                   ' Print some spaces to format the output 

  DO WHILE LEN(Binary$) <> bits%   ' Format to number of bits% long
    Binary$ = "0" + Binary$        ' Concatenate a 0 onto Binary$ string
  LOOP

  PRINT Binary$;                   ' Binary result, semicolon supresses carriage return
  'PRINT Binary$                   ' Binary result, implicit carriage return

END SUB  


'-------------------------------------------------------------------------------------------
' GetMazeSettigs  input parameters for maze generation
'-------------------------------------------------------------------------------------------

SUB GetMazeSettigs
   LOCAL STRING loopyn$
   LOCAL STRING biasn$

'Input Maze WIDTH
do
 print @(10,100) "                           ";
 print @(10,100) " ";
 input "Enter map width: (10-40) ", WIDTH% 
 WIDTH% = (WIDTH% \ 2)
 if WIDTH% > 4 and WIDTH% < 21 then exit do
loop

'Input Maze LENGTH
do
 print @(10,120) "                           ";
 print @(10,120) " ";
 input "Enter map height: (10-40) ", LENTH%
 LENTH% = (LENTH% \ 2)
 if LENTH% > 4 and LENTH% < 21 then exit do
loop

' Horizontal/Vertical bias: .5 = even, lower = more horizontal,  higher = more vertical
'Inpute Maze BIAS
print @(10,140) "The maze can be biased horizontally or vertically. 50=even bias, lower=more horizontal.";
do
 print @(10,160) "                           ";
 print @(10,160) " ";
 input "Enter bias: (1-99) ", BIAS
 if BIAS > 0 and BIAS < 100 then exit do
loop
BIAS = BIAS / 100

' Allow FunkyBias
print @(10,180) "Experimental variable Bias."
FUNKYBIAS% = 0
do
 print @(10,200) "                           ";
 print @(10,200) " ";
 input "Funky Bias: (y/n) "; biasn$
 if biasn$="Y" or biasn$="y" or biasn$="N" or biasn$="n" then exit do
loop
if biasn$="Y" or biasn$="y" then FUNKYBIAS% = 1

' Allow loops. 0 = NO loops, 1 = loops permited
print @(10,180) "Loops can be allowed in the maze, although not guaranteed."
LOOPS% = 0
do
 print @(10,200) "                           ";
 print @(10,200) " ";
 input "Allow loops: (y/n) "; loopyn$
 if loopyn$="Y" or loopyn$="y" or loopyn$="N" or loopyn$="n" then exit do
loop
if loopyn$="Y" or loopyn$="y" then LOOPS% = 1

END SUB

'-------------------------------------------------------------------------------------------
' END OF PROGRAM
'-------------------------------------------------------------------------------------------

